/* 
 * Solution of Exercise 9 - Numerics of Transport in Porous Media
 */

#ifndef GODUNOV_ASSEMBLER_HH
#define GODUNOV_ASSEMBLER_HH

#include "Vector.hh"
#include "grid.hh"
#include "spatial_parameters.hh"
#include "gw_bc.hh"
#include "godunov_utilities.hh"

// Default values for the precompiler macros

// default matrix is full matrix
#ifndef MATRIXTYPE
#define MATRIXTYPE DenseMatrix
#endif

class AdvectionDispersionAssembler
{
public:
  typedef PressureVelocityConverter AdvectionField;
  typedef ConstantDispersion DispersionField;

private:
  const Grid & grid_;
  std::vector<size_t> increments;
  std::vector<size_t> cells;

  const BoundaryConditions & boundary_conditions_;
  const DispersionField & dispersion_;
  const AdvectionField & advection_;
  const Sources & sources;




  /*
   *  An implementation of the MinMod Limiter
   */
  double get_rea_slope(
					   const Vector & x, 
					   const size_t index, 
					   Element & e,
					   const unsigned int dim, 
					   const double h
					   )
  {

    std::vector<double> n(2,0.0);
	
    for( FaceIterator fit = e.begin()
		   ; fit != e.end()
		   ; ++fit
		 )
	  {
		if((*fit).normal_component() != dim)
		  continue;
		
		int nindex;
		if((*fit).is_boundary())
		  nindex = index;
		else 
		  nindex = fit.out_id();

		const int inc = (*fit).normal_direction() > 0 ? 1 : 0;
		n[inc] = (double)(*fit).normal_direction() * (x[nindex] - x[index]) / h;
	  }
	
    const double & a = n[0];
    const double & b = n[1];
    if(a*b < 0)
      return 0;
    else if(std::abs(a) < std::abs(b))
      return a;
    else 
      return b;
  }
  
public:
  typedef MATRIXTYPE Matrix;

  AdvectionDispersionAssembler(
							   const Grid & _grid, 
							   const DispersionField & _dispersion,
							   const AdvectionField & _advection,
							   const BoundaryConditions & _boundary_conditions, 
                               const Sources & _sources
							   )
	: 
	grid_(_grid),
	increments(Grid::dimensions),
	cells(grid_.cells()),
	boundary_conditions_(_boundary_conditions),
	dispersion_(_dispersion),
	advection_(_advection),
	sources(_sources)
  {
    increments[0]=1;
    increments[1]=cells[0];
    if(Grid::dimensions==3)
      increments[2] = cells[1] * cells[0];
  }
  


  void assemble(
				Matrix & A, 
				Vector & b, 
				const Vector & x, 
				const double tau, 
				const double theta,
				const bool upwinding = true
				)
  {
	
    // loop over grid elements
    for( ElementIterator it = grid_.begin()
		   ; it != grid_.end()
		   ; ++it
		 )
	  {
		const Vector barycenter = (*it).barycenter();

		const size_t in_id = it.id();

		// loop over element faces
		for( FaceIterator fit = (*it).begin()
			   ; fit != (*it).end()
			   ; ++fit
			 )
		  {
			
			const double a = (*fit).area();
			
			const unsigned int normal_component = (*fit).normal_component();
			const int normal_direction = (*fit).normal_direction();
			
			Vector position = barycenter;
			const double h_2 = (*fit).dist_to_barycenter();
			position[normal_component] += (double)normal_direction * h_2;
			
			// if face has neighbor
			if(!(*fit).is_boundary()){
			  
			  const size_t out_id = fit.out_id();
			  
			  if(theta==0. && x[in_id]==0. && x[out_id]==0.)
				continue;
			  
			  // Dispersion
			  const double D = dispersion_(in_id,out_id,position,normal_component,normal_direction);
			  const double coeff = a / h_2 * 0.5 * D;
			  
			  A(in_id,in_id) += coeff * theta * tau;
			  A(in_id,out_id) += - coeff * theta * tau;
			  
			  b[in_id] -= coeff * (1.0 - theta) * x[in_id] * tau;
			  b[in_id] -= - coeff * (1.0 - theta) * x[out_id] * tau;
			  
			  // Advection multipied with normal vector
			  double ADV = advection_(
									  in_id,
									  out_id,
									  position,
									  normal_component,
									  normal_direction
									  );
			  
			  if(!upwinding)
				{
				  const double coeff = a * ADV * tau;
				  A(in_id,in_id) += coeff*0.5;
				  A(in_id,out_id) += coeff*0.5;
				}
			  else
				{
				  const double h = h_2*2.0;
				  const double coeff = a * ADV * tau;
				  
				  if(ADV>=0)
					{
					  A(in_id,in_id) += coeff * theta;
					  b[in_id] -= coeff * (1.-theta) * x[in_id];
					}
				  else
					{
					  A(in_id,out_id) += coeff * theta;
					  b[in_id] -= coeff * (1.-theta) * x[out_id]; 
					}
				}
			}
			// if face is on domain boundary
			else
			  {
				Vector barycenter = (*it).barycenter();
				const double h_2 = (*it).extent()[normal_component] * 0.5;
				barycenter[normal_component] += (double)normal_direction * h_2;
				
				// if dirichlet condition
				double dirichlet_value=0;
				if(boundary_conditions_.is_dirichlet(barycenter, &dirichlet_value))
				  {
				  
					// Dispersion
					const double D = dispersion_(in_id,in_id,position,normal_component,normal_direction);
					const double coeff = a / h_2 * D * tau;
				  
					A(in_id,in_id) += coeff * theta;
					b[in_id] -= coeff * (1.0-theta) * x[in_id];
					b[in_id] += coeff * dirichlet_value ;
					
					// Advectionn
					const double ADV = advection_(in_id,in_id,position,normal_component,normal_direction);
					
					if(!upwinding)
					  {
						const double coeff = a * ADV * tau;
						b[in_id] -= coeff * dirichlet_value;
					  }
					else
					  {
						const double coeff = a * ADV * tau;
						const double h = h_2*2.0;
	      
						if(ADV>=0)
						  {
							A(in_id,in_id) += coeff * theta;
							b[in_id] -= coeff * (1.-theta) * x[in_id];
						  }
						else
						  b[in_id] -= coeff * dirichlet_value;
						
					  }
				  }
				// if neumann condition
				else
				  {
					double neumann_value = boundary_conditions_.get_neumann(barycenter);
					b[in_id] -= a * neumann_value * tau;
				  }
				
			  } // if boundary
			
		  }// fit
		
		const double volume = (*it).volume();
		const Vector center = (*it).barycenter();
		const double id = it.id();
		
		// assemble source / sink term
		double source = sources(center);
		if(source!=0.0)
		  b[id] += source*volume;
		
		// assemble mass term
		b[id] += x[it.id()] * volume;
		A(id,id) += volume ;
		
	  }// it
    
  }
  







  void assemble_explicit(
						 Matrix & A, 
						 Vector & b, 
						 const Vector & x, 
						 const double tau,
						 bool slope
						 )
  {
	// INSERT YOUR CODE HERE TO IMPLEMENT THE TIME EXPLICIT FV-SCHEME WITH SLOPE LIMITERS
  }
  
};

#endif // GODUNOV_ASSEMBLER_HH
